<h1 class="logo"><a href="{{ pathto(master_doc) }}">{{ project }}</a></h1>

<p>{{ theme_description }}</p>
<p>Developed by Stefan Kasberger at AUSSDA - The Austrian Social Science Data Archive.</p>

<p>
  <a class="badge" href="https://travis-ci.com/gdcc/pyDataverse">
    <img alt="https://travis-ci.com/gdcc/pyDataverse.svg?branch=master" src="https://travis-ci.com/gdcc/pyDataverse.svg?branch=master">
  </a>
</p>

<iframe src="https://ghbtns.com/github-btn.html?user=gdcc&repo=pyDataverse&type=watch&count=true&size=large"
allowtransparency="true" frameborder="0" scrolling="0" width="200px" height="35px"></iframe>

<p><a href="https://twitter.com/theaussda" class="twitter-follow-button" data-show-count="false">Follow @theaussda</a> <script>!function(d,s,id){var js,fjs=d.getElementsByTagName(s)[0],p=/^http:/.test(d.location)?'http':'https';if(!d.getElementById(id)){js=d.createElement(s);js.id=id;js.src=p+'://platform.twitter.com/widgets.js';fjs.parentNode.insertBefore(js,fjs);}}(document, 'script', 'twitter-wjs');</script></p>
