"""Find out more at https://github.com/GDCC/pyDataverse.

Copyright 2019 Stefan Kasberger

Licensed under the MIT License.
"""

from __future__ import absolute_import


__author__ = "Stefan Kasberger"
__email__ = "stefan.kasberger@univie.ac.at"
__copyright__ = "Copyright (c) 2019 Stefan Kasberger"
__license__ = "MIT License"
__version__ = "0.3.5"
__url__ = "https://github.com/GDCC/pyDataverse"
__download_url__ = "https://pypi.python.org/pypi/pyDataverse"
__description__ = "A Python module for Dataverse."
